import inspect
import os

os.environ['PYGAME_HIDE_SUPPORT_PROMPT'] = 'hide'

from cmu_graphics.shape_logic import TRANSLATED_KEY_NAMES, _ShapeMetaclass
from cmu_graphics import shape_logic


class Signal:
    def __init__(self):
        self.receivers = []

    def connect(self, receiver):
        self.receivers.append(receiver)

    def send_robust(self, *args, **kwargs):
        for receiver in self.receivers:
            try:
                receiver(*args, **kwargs)
            except Exception:
                print('\nAn error occurred in a signal receiver')
                import traceback

                traceback.print_exc()


pygameEvent = Signal()
onStepEvent = Signal()
onMainLoopEvent = Signal()

EPSILON = 10e-7


def almostEqual(x, y, epsilon=EPSILON):
    return abs(x - y) <= epsilon


def rounded(d):
    sign = 1 if (d >= 0) else -1
    d = abs(d)
    n = int(d)
    if d - n >= 0.5:
        n += 1
    return sign * n


def round(*args):
    raise Exception(
        t(
            "Use our rounded(n) instead of Python 3's round(n)\n  Python 3's round(n) does not work as one might expect!\n  If you still want Python 3's round, use pythonRound"
        )
    )


def dsin(angle):
    return math.sin(math.radians(angle))


def dcos(angle):
    return math.cos(math.radians(angle))


def setLanguage(language):
    sli.setLanguage(language)


_print = print


def print(*args, **kwargs):
    return _print(*args, **kwargs)


def Robot(*args, **kwargs):
    raise NotImplementedError()


def assertEqual(*args, **kwargs):
    raise NotImplementedError()


class Shape(object, metaclass=_ShapeMetaclass):
    # This represents the attributes and methods handled by JS that the user
    # can call/get/set
    _js_attrs = {
        'left',
        'right',
        'top',
        'bottom',
        'centerX',
        'centerY',
        'width',
        'height',
        'fill',
        'opacity',
        'border',
        'borderWidth',
        'dashes',
        'align',
        'rotateAngle',
        'visible',
        'group',  # Not sure if this should be documented?
        'toBack',
        'toFront',
        'contains',
        'hits',
        'containsShape',
        'hitsShape',
        'rotate',
    }

    # This represents the valid keyword arguments passed to the constructor
    _init_attrs = {
        'fill',
        'border',
        'borderWidth',
        'opacity',
        'rotateAngle',
        'dashes',
        'align',
        'visible',
        'db',
    }

    def __init__(self, clsName, argNames, args, kwargs):
        if app is not None and app._app._isMvc:
            shapeName = self.__class__.__name__
            raise NotImplementedError(
                f'Whoops! {shapeName} objects are not available in CS3 Mode. Did you want draw{shapeName}?'
            )

        global SHAPES_CREATED
        SHAPES_CREATED += 1

        isMvc = False
        if 'isMvc' in kwargs:
            isMvc = kwargs['isMvc']
            del kwargs['isMvc']

        for attr in list(kwargs.keys()):
            en_attr = toEnglish(attr, 'shape-attr')
            if attr != en_attr and en_attr is not None:
                kwargs[en_attr] = kwargs[attr]
                del kwargs[attr]

            if en_attr not in self._init_attrs:
                raise Exception(
                    t(
                        "{{error}}: {{callSpec}} got an unexpected keyword argument '{{arg}}'",
                        {
                            'error': t('TypeError'),
                            'callSpec': t(clsName) + '()',
                            'arg': attr,
                        },
                    )
                )

        self._shape = slInitShape(clsName, argNames, args, kwargs, isMvc)
        self._shape.studentShape = self

    def __setattr__(self, attr, val):
        if attr[0] == '_':
            self.__dict__[attr] = val
        else:
            en_attr = toEnglish(attr, 'shape-attr')
            if en_attr in self._js_attrs:
                sli.slSetWithTypeCheck(self._shape, en_attr, val)
            else:
                self.__dict__[attr] = val
        return val

    def __getattr__(self, attr):
        if attr[0] == '_':
            return self.__dict__[attr]

        en_attr = toEnglish(attr, 'shape-attr')
        if en_attr in self._js_attrs:
            return slGet(self._shape, en_attr)
        else:
            return self.__dict__[attr]

    def __repr__(self):
        return self._shape._toString()


class Rect(Shape):
    def __init__(self, *args, **kwargs):
        super().__init__('Rect', ['left', 'top', 'width', 'height'], args, kwargs)


class Image(Shape):
    _js_attrs = Shape._js_attrs | {'url'}
    _init_attrs = Shape._init_attrs | {'height', 'width'}

    def __init__(self, *args, **kwargs):
        super().__init__('Image', ['url', 'left', 'top'], args, kwargs)


class Oval(Shape):
    def __init__(self, *args, **kwargs):
        super().__init__(
            'Oval', ['centerX', 'centerY', 'width', 'height'], args, kwargs
        )


class Circle(Shape):
    _js_attrs = Shape._js_attrs | {'radius'}

    def __init__(self, *args, **kwargs):
        super().__init__('Circle', ['centerX', 'centerY', 'radius'], args, kwargs)


class RegularPolygon(Shape):
    _js_attrs = Shape._js_attrs | {'radius', 'points'}

    def __init__(self, *args, **kwargs):
        super().__init__(
            'RegularPolygon', ['centerX', 'centerY', 'radius', 'points'], args, kwargs
        )


class Star(Shape):
    _js_attrs = Shape._js_attrs | {'radius', 'points', 'roundness'}
    _init_attrs = Shape._init_attrs | {'roundness'}

    def __init__(self, *args, **kwargs):
        super().__init__(
            'Star', ['centerX', 'centerY', 'radius', 'points'], args, kwargs
        )


class Line(Shape):
    _js_attrs = Shape._js_attrs | {
        'x1',
        'y1',
        'x2',
        'y2',
        'lineWidth',
        'arrowStart',
        'arrowEnd',
    }
    _init_attrs = (Shape._init_attrs | {'lineWidth', 'arrowStart', 'arrowEnd'}) - {
        'align',
        'border',
        'borderWidth',
    }

    def __init__(self, *args, **kwargs):
        super().__init__('Line', ['x1', 'y1', 'x2', 'y2'], args, kwargs)


class Polygon(Shape):
    _js_attrs = Shape._js_attrs | {'addPoint', 'pointList'}
    _init_attrs = Shape._init_attrs - {'align'}

    def __init__(self, *args, **kwargs):
        super().__init__('Polygon', ['initialPoints'], [args], kwargs)


class Arc(Shape):
    _js_attrs = Shape._js_attrs | {'startAngle', 'sweepAngle'}
    _init_attrs = Shape._init_attrs - {'align'}

    def __init__(self, *args, **kwargs):
        super().__init__(
            'Arc',
            ['centerX', 'centerY', 'width', 'height', 'startAngle', 'sweepAngle'],
            args,
            kwargs,
        )


class Label(Shape):
    _js_attrs = Shape._js_attrs | {'value', 'font', 'size', 'bold', 'italic'}
    _init_attrs = (Shape._init_attrs | {'bold', 'italic', 'size', 'font'}) - {'dashes'}

    def __init__(self, *args, **kwargs):
        super().__init__('Label', ['value', 'centerX', 'centerY'], args, kwargs)


class Group(Shape):
    _js_attrs = Shape._js_attrs | {
        'children',
        'add',
        'clear',
        'remove',
        'hitTest',
        # these attributes are not pass-through, so will throw an error if used
        'arrowEnd',
        'arrowStart',
        'url',
        'radius',
        'points',
        'roundness',
        'x1',
        'y1',
        'x2',
        'y2',
        'lineWidth',
        'startAngle',
        'sweepAngle',
        'value',
        'font',
        'size',
        'bold',
        'italic',
    }
    _init_attrs = {'visible', 'db'}

    def __init__(self, *args, **kwargs):
        if app is not None and app._app._isMvc:
            raise NotImplementedError(
                'Whoops! Group objects are not available in CS3 Mode.'
            )
        super().__init__('Group', [], [], kwargs)
        for shape in args:
            self.add(shape)

    def __iter__(self):
        return iter(self._shape)

    def __len__(self):
        return len(self._shape._shapes)


class Sound(object):
    number_of_sounds = 0

    def __init__(self, url):
        if not pygame.mixer.get_init():
            pygame.mixer.init()
            pygame.mixer.set_num_channels(1)

        if not isinstance(url, str):
            callSpec = '{className}.{attr}'.format(className=t('Sound'), attr=t('url'))
            err = t(
                '{{error}}: {{callSpec}} should be {{typeName}} (but {{value}} is of type {{valueType}})',
                {
                    'error': t('TypeError'),
                    'callSpec': callSpec,
                    'typeName': 'string',
                    'value': repr(url),
                    'valueType': type(url).__name__,
                },
            )
            raise Exception(err)

        Sound.number_of_sounds += 1
        if pygame.mixer.get_num_channels() == Sound.number_of_sounds:
            pygame.mixer.set_num_channels(Sound.number_of_sounds * 2)

        if url.startswith('file://'):
            url = url.split('//')[-1]

        if url.startswith('http'):
            try:
                response = webrequest.get(url)
                self.sound = io.BytesIO(response.read())
            except Exception:
                raise Exception('Failed to load sound data')

        elif hasattr(__main__, '__file__'):
            self.sound = os.path.abspath(os.path.join(__main__.__file__, '..', url))
        else:
            self.sound = os.path.abspath(os.path.join(os.getcwd(), url))

        self.sound = pygame.mixer.Sound(self.sound)
        self.channel = None

    def play(self, **kwargs):
        default_kwargs = {'loop': False, 'restart': False}

        for keyword in kwargs:
            english_keyword = toEnglish(keyword, 'shape-attr')
            if english_keyword not in default_kwargs:
                raise Exception(
                    "TypeError: %s.%s() got an unexpected keyword argument '%s'"
                    % (t('Sound'), t('play'), keyword)
                )
            default_kwargs[english_keyword] = kwargs[keyword]

        loop = default_kwargs['loop']
        restart = default_kwargs['restart']

        if not isinstance(loop, bool):
            raise Exception(
                'The loop argument to Sound.play must be True or False, got '
                + repr(loop)
            )
        if not isinstance(restart, bool):
            raise Exception(
                'The restart argument to Sound.play must be True or False, got '
                + repr(restart)
            )

        loop = -1 if loop else 0
        if not self.channel or not self.channel.get_busy():
            self.channel = self.sound.play(loops=loop)
        elif restart and self.channel.get_sound() != self.sound:
            self.channel = self.sound.play(loops=loop)
        elif restart:
            self.channel.stop()
            self.channel = self.sound.play(loops=loop)
        else:
            self.channel.unpause()

    def pause(self):
        self.channel.pause()

    def setVolume(self, volume: float):
        """
        Volume in the range of 0.0 to 1.0 (inclusive)\n
        If value < 0.0, the volume will not be changed\n
        If value > 1.0, the volume will be set to 1.0
        """
        self.sound.set_volume(volume)

    def getVolume(self):
        """
        Returns the volume (range: 0.0 - 1.0 (inclusive))
        """
        return self.sound.get_volume()


SHAPES = [
    Arc,
    Circle,
    Image,
    Label,
    Line,
    Oval,
    Polygon,
    Rect,
    RegularPolygon,
    Star,
]

enEventHandlerNamesMinusOnAppStart = [
    'onKeyPress',
    'onKeyHold',
    'onKeyRelease',
    'onMousePress',
    'onMouseDrag',
    'onMouseRelease',
    'onMouseMove',
    'onResize',
    'onStep',
    'redrawAll',
]

eventHandlerTranslationsMinusOnAppStart = list(
        enEventHandlerNamesMinusOnAppStart
) # Make a copy

onAppStartTranslations = ['onAppStart']

for language, translations in shape_logic.TRANSLATED_USER_FUNCTION_NAMES.items():
    if language == 'keys':
        continue
    for appFnName in enEventHandlerNamesMinusOnAppStart:
        for appFnNameTranslation in translations.get(appFnName, []):
            if appFnNameTranslation not in eventHandlerTranslationsMinusOnAppStart:
                eventHandlerTranslationsMinusOnAppStart.append(appFnNameTranslation)

    for onAppStartTranslation in translations.get('onAppStart', []):
        if onAppStartTranslation not in onAppStartTranslations:
            onAppStartTranslations.append(onAppStartTranslation)


class NoMvc:
    def __enter__(self):
        self.oldMvc = app._app._isMvc
        app._app._isMvc = False

    def __exit__(self, excType, excValue, tb):
        app._app._isMvc = self.oldMvc


def makeDrawFn(shape):
    def drawFn(*args, **kwargs):
        if not app._app._isMvc:
            raise Exception(
                f'You called draw{shape.__name__} (a CS3 Mode function) outside of redrawAll.'
            )
        if not app._app.inRedrawAll:
            raise MvcException('Cannot draw (modify the view) outside of redrawAll')
        with NoMvc():
            kwargs['isMvc'] = True
            shape(*args, **kwargs)

    return drawFn


def makeInvisibleConstructor(shape):
    def constructor(*args, **kwargs):
        if not app._app._isMvc:
            raise Exception(
                f'You called {shape.__name__}Shape (a CS3 Mode function) outside of CS3 Mode. To run your app in CS3 Mode, use runApp().'
            )
        with NoMvc():
            result = shape(*args, **kwargs)
        result.visible = False
        return result

    return constructor


def createDrawingFunctions():
    g = globals()
    for shape in SHAPES:
        shapeName = shape.__name__
        if shapeName == 'Group':
            continue
        g['draw' + shapeName] = makeDrawFn(shape)
        g[shapeName + 'Shape'] = makeInvisibleConstructor(shape)


createDrawingFunctions()


class KeyName(str):
    def __init__(self, baseKey):
        self.__dict__['accentCombinations'] = accentCombinations(str(self))

    def __eq__(self, other):
        return other in self.accentCombinations

    def __setattr__(self, attr, value):
        raise AttributeError(f"'str' object has no attribute '{attr}'")


def translateKeyName(keyName, originalLanguage):
    if originalLanguage not in TRANSLATED_KEY_NAMES:
        return keyName
    return KeyName(TRANSLATED_KEY_NAMES[originalLanguage].get(keyName, keyName))


def cleanAndClose():
    try:
        app._app.callUserFn('onAppStop', (), redraw=False)
    except Exception:
        pass
    os._exit(0)


def _safeMethod(appMethod):
    def m(*args, **kwargs):
        app = args[0]
        try:
            return appMethod(*args, **kwargs)
        except Exception:
            sys.excepthook(*sys.exc_info())
            app.stop()
            if app._running:
                app.drawErrorScreen()
            else:
                cleanAndClose()

    return m


# Based on Lukas Peraza's pygame framework
# https://github.com/LBPeraza/Pygame-Asteroids
class App(object):
    def printFullTracebacks(self):
        shape_logic.printFullTracebacks()

    def getScreenshot(self, path):
        with DRAWING_LOCK:
            pygame.image.save(self._screen, path)

    def quit(self):
        self._running = False

    def getPosArgCount(self, fn):
        fn_code = fn.__code__
        pos_count = fn_code.co_argcount
        arg_names = fn_code.co_varnames
        return len(arg_names[:pos_count])

    def usesControl(self, fn):
        fn_code = fn.__code__
        return 'control' in fn_code.co_consts

    def getFnNameAndLanguage(self, enFnName, useActiveScreen):
        if enFnName in self.userGlobals:
            return enFnName, 'en'
        
        if app._app._initialScreen is not None and useActiveScreen:
            screenFnName = f'{self.activeScreen}_{enFnName}'
            if screenFnName in self.userGlobals:
                return screenFnName, 'en'

        for language, translations in shape_logic.TRANSLATED_USER_FUNCTION_NAMES.items():
            if language == 'keys':
                continue

            for fnTranslation in translations.get(enFnName, []):
                # We always use the non-screen version of the function first if
                # it exists in the code. This is ok, because non-screen
                # functions are not allowed when you're running an app with
                # screens (except for onAppStart)
                if fnTranslation in self.userGlobals:
                    return fnTranslation, language

                if app._app._initialScreen is not None and useActiveScreen:
                    screenFnName = f'{self.activeScreen}_{fnTranslation}'
                    if screenFnName in self.userGlobals:
                        return screenFnName, language

        return None, None

    def translateEventHandlerArgs(self, baseFnName, language, args):
        if baseFnName == 'onKeyHold':
            args = ([translateKeyName(x, language) for x in args[0]],)
        elif baseFnName in ('onKeyPress', 'onKeyRelease'):
            args = (translateKeyName(args[0], language), args[1])

        return args

    def getEventHandlerArgs(self, baseFnName, language, fn, args, kwargs):
        if language != 'en':
            args = self.translateEventHandlerArgs(baseFnName, language, args)

        if self._isMvc:
            args = (self._wrapper,) + args

        if baseFnName in (
            'onKeyPress',
            'onKeyRelease',
            'onKeyHold',
            'onMousePress',
            'onMouseRelease',
            'onMouseDrag',
        ):
            if self.getPosArgCount(fn) < len(args):
                args = args[:-1]
            elif (
                baseFnName in ('onKeyPress', 'onKeyRelease', 'onKeyHold')
                and self.shouldPrintCtrlWarning
                and self.usesControl(fn)
            ):
                print('INFO: To use the control key in your app without')
                print('enabling the inspector, set app.inspectorEnabled')
                print('to False. To stop this message from printing,')
                print('set app.inspectorEnabled to True.')
                self.shouldPrintCtrlWarning = False

        return args, kwargs

    @_safeMethod
    def callUserFn(self, baseFnName, args, kwargs=None, redraw=True, useActiveScreen=True):
        if kwargs is None:
            kwargs = dict()

        fnName, language = self.getFnNameAndLanguage(baseFnName, useActiveScreen)
        if fnName is None:
            return

        fn = self.userGlobals[fnName]
        args, kwargs = self.getEventHandlerArgs(baseFnName, language, fn, args, kwargs)

        fn(*args, **kwargs)

        if redraw and self._isMvc and baseFnName != 'redrawAll':
            self.redrawAllWrapper()

    def redrawAllWrapper(self):
        self.group.clear()

        self.inRedrawAll = True
        self.callUserFn('redrawAll', ())
        self.inRedrawAll = False

    @staticmethod
    def getKey(keyCode, modifierMask):
        keyNameMap = {
            pygame.K_TAB: 'tab',
            pygame.K_RETURN: 'enter',
            pygame.K_BACKSPACE: 'backspace',
            pygame.K_DELETE: 'delete',
            pygame.K_ESCAPE: 'escape',
            pygame.K_SPACE: 'space',
            pygame.K_RIGHT: 'right',
            pygame.K_LEFT: 'left',
            pygame.K_UP: 'up',
            pygame.K_DOWN: 'down',
            pygame.K_RCTRL: 'ctrl',
            pygame.K_LCTRL: 'ctrl',
        }

        shiftMap = {
            '1': '!',
            '2': '@',
            '3': '#',
            '4': '$',
            '5': '%',
            '6': '^',
            '7': '&',
            '8': '*',
            '9': '(',
            '0': ')',
            '[': '{',
            ']': '}',
            '/': '?',
            '=': '+',
            '\\': '|',
            "'": '"',
            ',': '<',
            '.': '>',
            '-': '_',
            ';': ':',
            '`': '~',
        }

        # Punctuation, numbers, and letters
        if 33 < keyCode < 127:
            key = chr(keyCode)
            if modifierMask & pygame.KMOD_SHIFT:
                key = shiftMap.get(key, key).upper()
            return key
        return keyNameMap.get(keyCode, None)

    def drawErrorScreen(self):
        cairo_surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, self.width, self.height)
        ctx = cairo.Context(cairo_surface)

        with NoMvc():
            Rect(0, 0, self.width, self.height, fill=None, border='red', borderWidth=2)
            Rect(
                10,
                self.height - 60,
                self.width - 20,
                50,
                fill='white',
                border='red',
                borderWidth=4,
            )
            Label(
                'Exception! App Stopped!',
                self.width / 2,
                self.height - 45,
                size=12,
                bold=True,
                font='Arial',
                fill='red',
            )
            Label(
                'See console for details',
                self.width / 2,
                self.height - 25,
                size=12,
                bold=True,
                font='Arial',
                fill='red',
            )

        self.redrawAll(self._screen, cairo_surface, ctx)

    def getModifiers(self, modifierMask):
        modifiers = list()
        if modifierMask & pygame.KMOD_SHIFT:
            modifiers.append('shift')
        if modifierMask & pygame.KMOD_CTRL:
            modifiers.append('control')
        if modifierMask & pygame.KMOD_META:
            modifiers.append('meta')
        return modifiers

    def handleKeyPress(self, keyCode, modifierMask):
        self._modifiers = self.getModifiers(modifierMask)
        key = App.getKey(keyCode, modifierMask)

        if key is None:
            return
        if key == 'ctrl':
            self.isCtrlKeyDown = True
            return
        if key == 'space' and (modifierMask & pygame.KMOD_SHIFT):
            self.paused = not self.paused
            return

        self._allKeysDown.add(key)

        modifiers = self.getModifiers(modifierMask)
        self.callUserFn('onKeyPress', (key, modifiers))

    def handleKeyRelease(self, keyCode, modifierMask):
        self._modifiers = self.getModifiers(modifierMask)
        key = App.getKey(keyCode, modifierMask)

        if key is None:
            return
        if key == 'ctrl':
            self.isCtrlKeyDown = False
            return
        if key.upper() in self._allKeysDown:
            self._allKeysDown.remove(key.upper())
        if key.lower() in self._allKeysDown:
            self._allKeysDown.remove(key.lower())

        modifiers = self.getModifiers(modifierMask)
        self.callUserFn('onKeyRelease', (key, modifiers))

    def redrawAll(self, screen, cairo_surface, ctx):
        shape = shape_logic.Rect(
            {
                'noGroup': True,
                'top': 0,
                'left': 0,
                'width': self.width,
                'height': self.height,
                'fill': self.background or 'white',
            }
        )
        shape.draw(ctx)

        ctx.save()
        try:
            self._tlg._shape.draw(ctx)
        finally:
            ctx.restore()

        ctx.save()
        try:
            if self.shouldDrawInspector():
                self.inspector.draw(ctx)
        finally:
            ctx.restore()

        # Get the cairo buffer and convert it from BGRA to RGBA
        data_string = cairo_surface.get_data()

        # Create PyGame surface
        pygame_surface = pygame.image.frombuffer(
            data_string, (self.width, self.height), 'RGBA'
        )

        # Show PyGame surface
        screen.blit(pygame_surface, (0, 0))
        pygame.display.flip()

        self.frameworkRedrew = True

    def shouldDrawInspector(self):
        return self.inspectorEnabled and (
            self.paused or self.alwaysShowInspector or self.isCtrlKeyDown
        )

    def __init__(self):
        self.userGlobals = __main__.__dict__
        try:
            self.title, _ = os.path.splitext(
                os.path.basename(os.path.realpath(__main__.__file__))
            )
        except Exception:
            self.title = 'CMU CS Academy'

        self._width = 400
        self._height = 400
        self._allKeysDown = set()
        self._modifiers = set()
        self.background = None

        self._stepsPerSecond = 30

        self._tlg = Group()
        sli.setTopLevelGroup(self._tlg)

        self.paused = False
        self._stopped = False
        self._running = False
        self.textInputs = []

        self.inspector = shape_logic.Inspector(self)
        self._inspectorEnabled = True
        self.shouldPrintCtrlWarning = True
        self.alwaysShowInspector = False
        self.isCtrlKeyDown = False
        self.showFontWarnings = True

        self._isMvc = False
        self._initialScreen = None

    def get_group(self):
        return self._tlg

    def set_group(self, _):
        raise Exception('App.group is readonly')

    group = property(get_group, set_group)

    def get_stopped(self):
        return self._stopped

    def set_stopped(self, _):
        raise Exception('App.stopped is readonly')

    stopped = property(get_stopped, set_stopped)

    def getStepsPerSecond(self):
        return self._stepsPerSecond

    def setStepsPerSecond(self, value):
        shape_logic.checkNumber(sli.t('app'), 'stepsPerSecond', value, False)
        self._stepsPerSecond = value

    stepsPerSecond = property(getStepsPerSecond, setStepsPerSecond)

    def getBackground(self):
        return sli.slGetAppProperty('background')

    def setBackground(self, value):
        return sli.slSetAppProperty('background', value)

    background = property(getBackground, setBackground)

    def getMaxShapeCount(self):
        return sli.slGetAppProperty('maxShapeCount')

    def setMaxShapeCount(self, value):
        return sli.slSetAppProperty('maxShapeCount', value)

    maxShapeCount = property(getMaxShapeCount, setMaxShapeCount)

    def updateScreenSize(self):
        if self._running:
            self.updateScreen(True)

    def handleResize(self, newWidth, newHeight):
        self._width = newWidth
        self._height = newHeight
        self.updateScreen(False)

        # Redraw even if onResize is not present in the user's globals
        self.callUserFn('onResize', (), redraw=False)
        if self._isMvc:
            self.redrawAllWrapper()

    def handleSetActiveScreen(self, newScreen, redraw=True):
        self.activeScreen = newScreen
        self.callUserFn('onScreenActivate', (), redraw=redraw, useActiveScreen=True)

    def getLeft(self):
        return 0

    def setLeft(self, value):
        raise Exception('App.left is readonly')

    left = property(getLeft, setLeft)

    def getRight(self):
        return self._width

    def setRight(self, value):
        self._width = value
        self.updateScreenSize()

    right = property(getRight, setRight)

    def getTop(self):
        return 0

    def setTop(self, value):
        raise Exception(t('App.top is readonly'))

    top = property(getTop, setTop)

    def getBottom(self):
        return self._height

    def setBottom(self, value):
        self._height = value
        self.updateScreenSize()

    bottom = property(getBottom, setBottom)

    def getWidth(self):
        return self._width

    def setWidth(self, value):
        self._width = value
        self.updateScreenSize()

    width = property(getWidth, setWidth)

    def getHeight(self):
        return self._height

    def setHeight(self, value):
        self._height = value
        self.updateScreenSize()

    height = property(getHeight, setHeight)

    def get_inspectorEnabled(self):
        return self._inspectorEnabled

    def set_inspectorEnabled(self, value):
        self.shouldPrintCtrlWarning = False
        self._inspectorEnabled = value

    inspectorEnabled = property(get_inspectorEnabled, set_inspectorEnabled)

    def get_showFontWarnings(self):
        return shape_logic.SHOW_FONT_WARNINGS

    def set_showFontWarnings(self, value):
        shape_logic.SHOW_FONT_WARNINGS = value
    
    showFontWarnings = property(get_showFontWarnings, set_showFontWarnings)


    def stop(self):
        self._stopped = True

    def getTextInput(self, prompt='Enter some text'):
        if self.textInputs:
            return self.textInputs.pop(0)
        p = self.spawnModalProcess()
        packet = bytes(
            json.dumps({'title': self.title, 'prompt': prompt, 'getInput': True})
            + '\n',
            encoding='utf-8',
        )
        result, errors = p.communicate(packet)
        if p.returncode is not None and p.returncode != 0:
            print(errors.decode('utf-8'))
            raise Exception('Exception in getTextInput.')
        return result.decode('utf-8')

    def showMessage(self, prompt=''):
        p = self.spawnModalProcess()
        packet = bytes(
            json.dumps({'title': self.title, 'prompt': prompt, 'getInput': False})
            + '\n',
            encoding='utf-8',
        )
        unused_result, errors = p.communicate(packet)
        if p.returncode is not None and p.returncode != 0:
            print(errors.decode('utf-8'))
            raise Exception('Exception in showMessage.')

    def setTextInputs(self, *args):
        for arg in args:
            if not isinstance(arg, str):
                raise Exception(
                    'Arguments to setTextInputs must be strings. %r is not a string.'
                    % arg
                )
        self.textInputs = list(args)

    def spawnModalProcess(self):
        current_directory = os.path.dirname(os.path.realpath(__file__))
        modal_path = os.path.join(current_directory, 'modal.py')
        p = subprocess.Popen(
            [sys.executable, modal_path],
            stdout=subprocess.PIPE,
            stdin=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd=current_directory,
        )
        return p

    def updateScreen(self, newScreen):
        if newScreen:
            self._screen = pygame.display.set_mode(
                (self.width, self.height), pygame.RESIZABLE
            )
        self._cairo_surface = cairo.ImageSurface(
            cairo.FORMAT_ARGB32, self.width, self.height
        )
        self._ctx = cairo.Context(self._cairo_surface)

    @_safeMethod
    def run(self):
        pygame.init()
        pygame.display.set_caption(self.title)

        self._screen = None
        self.updateScreen(True)

        lastTick = 0
        self._running = True

        while self._running:
            sys.stdout.flush()
            with DRAWING_LOCK:
                had_event = False
                for event in pygame.event.get():
                    had_event = True
                    if not self.stopped:
                        if event.type == pygame.MOUSEBUTTONDOWN and event.button <= 3:
                            self.callUserFn(
                                'onMousePress', (*event.pos, event.button - 1)
                            )
                        elif event.type == pygame.MOUSEBUTTONUP and event.button <= 3:
                            self.callUserFn(
                                'onMouseRelease', (*event.pos, event.button - 1)
                            )
                        elif event.type == pygame.MOUSEMOTION:
                            if event.buttons == (0, 0, 0):
                                self.callUserFn('onMouseMove', event.pos)
                            else:
                                self.callUserFn(
                                    'onMouseDrag',
                                    (
                                        *event.pos,
                                        [i for i in range(3) if event.buttons[i] != 0],
                                    ),
                                )
                        elif event.type == pygame.KEYDOWN:
                            self.handleKeyPress(event.key, event.mod)
                        elif event.type == pygame.KEYUP:
                            self.handleKeyRelease(event.key, event.mod)
                        elif event.type == SET_ACTIVE_SCREEN:
                            self.handleSetActiveScreen(event.newScreen)
                        elif event.type == pygame.WINDOWSIZECHANGED:
                            self.handleResize(event.x, event.y)
                    if event.type == pygame.QUIT:
                        self._running = False
                    elif event.type == pygame.MOUSEMOTION:
                        self.inspector.setMousePosition(*event.pos)
                    elif event.type in (pygame.KEYDOWN, pygame.KEYUP):
                        key = App.getKey(event.key, event.mod)
                        if key == 'ctrl':
                            self.isCtrlKeyDown = event.type == pygame.KEYDOWN

                    pygameEvent.send_robust(event, self.callUserFn, self._wrapper)

                should_redraw = had_event

                msPassed = pygame.time.get_ticks() - lastTick
                if 1000 / self.stepsPerSecond - msPassed < 1:
                    lastTick = pygame.time.get_ticks()
                    if not (self.paused or self.stopped):
                        self.callUserFn('onStep', ())
                        if len(self._allKeysDown) > 0:
                            self.callUserFn(
                                'onKeyHold',
                                (list(self._allKeysDown), list(self._modifiers)),
                            )
                        onStepEvent.send_robust(self.callUserFn, self._wrapper)
                        should_redraw = True

                if should_redraw:
                    self.inspector.clearCache()
                    self.redrawAll(self._screen, self._cairo_surface, self._ctx)

                onMainLoopEvent.send_robust(msPassed, self.callUserFn, self._wrapper)

                pygame.time.wait(1)

        pygame.quit()
        cleanAndClose()


class MvcException(Exception):
    pass


class AppWrapper(object):
    readOnlyAttrs = set(
        [
            'bottom',
            'centerX',
            'centerY',
            'getTextInput',
            'showMessage',
            'left',
            'quit',
            'right',
            'run',
            'stop',
            'top',
            'setMaxShapeCount',
            'printFullTracebacks',
        ]
    )
    readWriteAttrs = set(
        [
            'height',
            'paused',
            'stepsPerSecond',
            'group',
            'title',
            'width',
            'background',
            'beatsPerMinute',
            'maxShapeCount',
            'inspectorEnabled',
            'showFontWarnings',
        ]
    )
    allAttrs = readOnlyAttrs | readWriteAttrs

    def __init__(self, app):
        self._app = app
        app._wrapper = self

    def __dir__(self):
        fields = set(AppWrapper.allAttrs)
        for field in self.__dict__:
            if field not in self._app.__dict__:
                fields.add(field)
        return sorted(fields)

    def __getattribute__(self, attr):
        attr = toEnglish(attr, 'app-attr')
        if attr == '_app' or attr not in AppWrapper.allAttrs:
            return super().__getattribute__(attr)
        return self._app.__getattribute__(attr)

    def __setattr__(self, attr, value):
        attr = toEnglish(attr, 'app-attr')
        if (attr != '_app') and (getattr(self._app, 'inRedrawAll', False)):
            raise MvcException(f'Cannot change app.{attr} in redrawAll')
        if attr in AppWrapper.readOnlyAttrs:
            raise Exception(f'app.{attr} is read-only')
        if attr in AppWrapper.readWriteAttrs:
            return self._app.__setattr__(attr, value)
        return super().__setattr__(attr, value)


def runApp(width=400, height=400, **kwargs):
    # If we didn't call runAppWithScreens
    if app._app._initialScreen is None:
        for appFnName in (eventHandlerTranslationsMinusOnAppStart + onAppStartTranslations):
            screenAppSuffix = f'_{appFnName}'
            for globalVarName in app._app.userGlobals:
                if globalVarName.endswith(screenAppSuffix):
                    raise Exception(
                        f'The name of your function "{globalVarName}" ends with "{screenAppSuffix}", which is only allowed if you are using "screens" in CS3 Mode. To run an app with screens, call runAppWithScreens() instead of runApp().'
                    )

    setupMvc()
    app.width = width
    app.height = height

    if SHAPES_CREATED > 1:
        raise Exception("""
****************************************************************************
Your code created a shape object (Rect, Oval, etc.) before calling runApp().

runApp (CS3 Mode) is not compatible with shape objects.

If you'd like to use CS3 Mode, please use drawing functions
(drawRect, drawOval, etc) in redrawAll.

Otherwise, please call cmu_graphics.run() in place of runApp.
****************************************************************************""")

    # Don't redraw on either of these calls to callUserFn, because we will
    # instead redraw below
    app._app.callUserFn('onAppStart', (), kwargs, redraw=False, useActiveScreen=False)            

    if app._app._initialScreen is not None:
        sortedGlobals = sorted(app._app.userGlobals)
        for onAppStartTranslation in onAppStartTranslations:
            for globalVarName in sortedGlobals:
                if globalVarName.endswith(f'_{onAppStartTranslation}'):
                    screenFn = app._app.userGlobals[globalVarName]
                    screenFn(app, **kwargs)

        setActiveScreen(app._app._initialScreen, fromRunApp=True)

    # Guarantee at least one call to redrawAll. For example, if your app
    # just sets up variables and draws a static image.
    app._app.redrawAllWrapper()

    run()


def setActiveScreen(screen, fromRunApp=False):
    if not app._app._isMvc:
        raise Exception(
            'You called setActiveScreen (a CS3 Mode function) outside of CS3 Mode. To run your app in CS3 Mode, use runApp() or runAppWithScreens().'
        )
    if (screen in [None, '']) or (not isinstance(screen, str)):
        raise Exception(f'{repr(screen)} is not a valid screen')
    redrawAllFnName = f'{screen}_redrawAll'
    if redrawAllFnName not in app._app.userGlobals:
        raise Exception(f'Screen {screen} requires {redrawAllFnName}()')
    if fromRunApp:
        app._app.handleSetActiveScreen(screen, redraw=False)
    else:
        pygame.event.post(pygame.event.Event(SET_ACTIVE_SCREEN, newScreen=screen))


def runAppWithScreens(initialScreen, *args, **kwargs):
    userGlobals = app._app.userGlobals

    for appFnName in eventHandlerTranslationsMinusOnAppStart:
            if appFnName in userGlobals:
                raise Exception(f'Do not define {appFnName} when using screens')
            
    app._app._isMvc = True
    app._app._initialScreen = initialScreen
    runApp(*args, **kwargs)


def getImageSize(url):
    imageData = shape_logic.loadImage(url)
    width, height = imageData['width'], imageData['height']
    return width, height


def setupMvc():
    app._app._isMvc = True
    app._app.inRedrawAll = False
    del app._app.userGlobals['app']
    AppWrapper.readWriteAttrs.remove('paused')
    AppWrapper.allAttrs.remove('paused')


def processArgs(fname, params, args):
    # Check for too many positional arguments
    if len(args) > len(params):
        argStr = 'argument' if len(params) == 1 else 'arguments'
        raise TypeError(
            f'{fname}() takes {len(params)} positional {argStr} but more were given'
        )

    # Check for not enough positional arguments
    if len(params) > len(args):
        missingCount = len(params) - len(args)
        argStr = 'argument' if missingCount == 1 else 'arguments'
        paramsStr = ', '.join([repr(param) for param in params[len(args) :]])
        raise TypeError(
            f'{fname}() missing {missingCount} required positional {argStr}: {paramsStr}'
        )


def eventHandlerRepeater(f):
    sig = inspect.signature(f)
    params = tuple(sig.parameters.keys())

    def g(*args):
        testParams = params
        if app._app._isMvc:
            testParams = ('app',) + testParams
        processArgs(f.__name__, testParams, args)
        if app._app._isMvc:
            args = args[1:]
        f(*args)

    return g


@eventHandlerRepeater
def onSteps(n):
    for _ in range(n):
        app._app.callUserFn('onStep', ())


@eventHandlerRepeater
def onKeyHolds(keys, n):
    assert isinstance(keys, list), t('keys must be a list')
    for _ in range(n):
        app._app.callUserFn('onKeyHold', (keys, []))


@eventHandlerRepeater
def onKeyPresses(key, n):
    for _ in range(n):
        app._app.callUserFn('onKeyPress', (key, []))


def loop():
    run()


def run():
    if not app._app._isMvc:
        for cs3ModeHandler in ['redrawAll']:
            if cs3ModeHandler in __main__.__dict__:
                raise Exception(
                    f"You defined the event handler {cs3ModeHandler} which works with CS3 mode, and then called cmu_graphics.run(), which doesn't work with CS3 mode. Did you mean to call runApp instead?"
                )

    global MAINLOOP_RUN
    MAINLOOP_RUN = True

    if not os.environ.get('CI', False):
        threading.Thread(target=CSAcademyConsole().interact).start()

    try:
        app._app.run()
    except KeyboardInterrupt:
        cleanAndClose()


from code import InteractiveConsole


class CSAcademyConsole(InteractiveConsole):
    def __init__(self):
        self.__class__.__name__ = 'CS Academy Console'
        __main__.__dict__['exit'] = lambda: cleanAndClose()
        super().__init__(
            locals=__main__.__dict__, filename='<%s>' % self.__class__.__name__
        )

    # Override the default error handling functions to avoid using our own
    # excepthook function
    def showsyntaxerror(self, filename=None):
        type, value, tb = sys.exc_info()
        sys.last_type = type
        sys.last_value = value
        sys.last_traceback = tb
        if filename and type is SyntaxError:
            # Work hard to stuff the correct filename in the exception
            try:
                msg, (dummy_filename, lineno, offset, line) = value.args
            except ValueError:
                # Not the format we expect; leave it alone
                pass
            else:
                # Stuff in the right filename
                value = SyntaxError(msg, (filename, lineno, offset, line))
                sys.last_value = value

        lines = traceback.format_exception_only(type, value)
        self.write(''.join(lines))

    def showtraceback(self):
        sys.last_type, sys.last_value, last_tb = ei = sys.exc_info()
        sys.last_traceback = last_tb
        try:
            lines = traceback.format_exception(ei[0], ei[1], last_tb.tb_next)
            self.write(''.join(lines))
        finally:
            last_tb = ei = None

    # Override interact so we can exit on EOF
    def interact(self):
        super().interact()
        cleanAndClose()


import sys
import io
from datetime import datetime
from datetime import timedelta
import json
import subprocess
from cmu_graphics.libs import webrequest
import __main__


UPDATE_CONFIG_FILE_PATH = os.path.join(
    os.path.dirname(os.path.realpath(__file__)),
    'meta',
    'updates.json',
)


def get_update_info():
    if os.path.exists(UPDATE_CONFIG_FILE_PATH):
        with open(UPDATE_CONFIG_FILE_PATH, 'r') as f:
            return json.loads(f.read())
    return {}


def save_update_info(update_info):
    with open(UPDATE_CONFIG_FILE_PATH, 'w') as f:
        f.write(json.dumps(update_info))


def check_for_update():
    try:
        update_info = get_update_info()

        current_directory = os.path.dirname(os.path.realpath(__file__))
        with open(os.path.join(current_directory, 'meta', 'version.txt')) as f:
            version = f.read().strip()

        last_attempt = None
        if 'last_attempt' in update_info:
            last_attempt = datetime.fromtimestamp(update_info['last_attempt'])

        if last_attempt is None or (datetime.now() - last_attempt > timedelta(days=1)):
            most_recent_version = (
                webrequest.get(
                    'https://s3.amazonaws.com/cmu-cs-academy.lib.prod/desktop-cmu-graphics/version.txt'
                )
                .read()
                .decode('ascii')
                .strip()
            )

            update_info['last_attempt'] = datetime.now().timestamp()
            update_info['most_recent_version'] = most_recent_version
            save_update_info(update_info)
        else:
            most_recent_version = update_info.get('most_recent_version', version)

        if most_recent_version > version:
            print(
                f'\n\nYou are running cmu-graphics version {version}, but a newer version {most_recent_version} is available.'
            )
            
            ### PYPI VERSION ###
            print('Run "pip install --upgrade cmu-graphics" to upgrade.')
            ### END PYPI VERSION ###
            print('\n\n')
    except Exception:
        pass


if 'CMU_GRAPHICS_NO_UPDATE' not in __main__.__dict__:
    check_for_update()


def print_debug_info():
    import platform

    current_directory = os.path.dirname(os.path.realpath(__file__))
    with open(os.path.join(current_directory, 'meta', 'version.txt')) as f:
        version = f.read().strip()
    print('=' * 80)
    print('CMU Graphics Version:', version)
    print('Platform:', sys.platform)
    print('Python Version:', '.'.join(platform.python_version_tuple()))
    print('Executable Path:', sys.executable)
    print('Python path:', sys.path)
    print('Working Directory:', current_directory)
    print('=' * 80)


if 'CMU_GRAPHICS_DEBUG' in __main__.__dict__:
    print_debug_info()

import math


### PYPI VERSION ###
import cairo

### END PYPI VERSION ###
from random import *
from cmu_graphics.utils import *
import atexit
import threading
import traceback

DRAWING_LOCK = threading.RLock()


### PYPI VERSION ###
import pygame
### END PYPI VERSION ###

sli = shape_logic.ShapeLogicInterface()
slInitShape = sli.slInitShape
slGet = sli.slGet
rgb = sli.rgb
gradient = sli.gradient
toEnglish = sli.toEnglish
accentCombinations = sli.accentCombinations
t = sli.t

SHAPES_CREATED = 0
MAINLOOP_RUN = False
SET_ACTIVE_SCREEN = pygame.event.custom_type()

# Checks to see if a user created shapes but did not call
# cmu_graphics.run()
def check_for_exit_without_run():
    global SHAPES_CREATED, MAINLOOP_RUN

    # The app's top level group is created even if the user creates no
    # shapes on their own
    if SHAPES_CREATED > 1 and not MAINLOOP_RUN:
        print("""
                         (
                    (    (
                    ((  (*(
                    (*( (*/
                    (**.***,
                    (***************((((((((((((((((
                    (********************************
                    (*******************************(
                    (*******************************(
                    (*******************************(
                    /*******************************(
                    (/******************(((((((     ((
                (*****(****************,
                /**********(************(
            ((***************(*********
                (*****(/*********(*****(
                    (**********/(/***(*/
                    (****************(
                        (/***********(
                            (*******(
                            (**(
""")
        print(
            ' ** To run your animation, add cmu_graphics.run() to the bottom of your file **\n'
        )


app = None
app = AppWrapper(App())
atexit.register(check_for_exit_without_run)



import __main__
import os
import sys


def check_for_shadowing():
    #   If the current working directory or the __main__ script's directory
    # are in sys.path and contain files with the same name as modules used
    # by cmu_graphics, complain.
    #   Note that there could be other directories in sys.path that contain
    # files that shadow installed modules that cmu_graphics imports. We are
    # currently not aware of a way to distinguish between "user" directories
    # in sys.path and "system" directories in sys.path, so we just consider
    # the two directories mentioned above.

    directories_to_look_for_shadowing = []

    # If the current working directory is in sys.path, consider it for shadowing
    if '' in sys.path:
        directories_to_look_for_shadowing.append(os.path.abspath('.'))

    # If the __main__ file's directory is in sys.path, consider it for shadowing
    # (When running in the Python 3.13 REPL, __main__.__file__ is not in sys.path.)
    if main_file := getattr(__main__, '__file__', None):
        main_directory = os.path.dirname(main_file)
        if main_directory in sys.path:
            directories_to_look_for_shadowing.append(main_directory)

    # disallowed_filenames was generated by running this code at the end of cmu_graphics.py:
    # import sys
    # print('\n'.join(sorted(f"'{x}'," for x in list(sys.modules.keys()) if '.' not in x)))
    disallowed_filenames = {
        f'{x}.py'
        for x in [
            '__future__',
            '__main__',
            '_abc',
            '_ast',
            '_bisect',
            '_blake2',
            '_bz2',
            '_codecs',
            '_collections',
            '_collections_abc',
            '_compression',
            '_datetime',
            '_decimal',
            '_frozen_importlib',
            '_frozen_importlib_external',
            '_functools',
            '_hashlib',
            '_heapq',
            '_imp',
            '_io',
            '_json',
            '_locale',
            '_lzma',
            '_opcode',
            '_operator',
            '_posixsubprocess',
            '_queue',
            '_random',
            '_scproxy',
            '_sha2',
            '_signal',
            '_sitebuiltins',
            '_socket',
            '_sre',
            '_ssl',
            '_stat',
            '_string',
            '_struct',
            '_thread',
            '_tokenize',
            '_uuid',
            '_warnings',
            '_weakref',
            '_weakrefset',
            'abc',
            'array',
            'ast',
            'atexit',
            'base64',
            'binascii',
            'bisect',
            'builtins',
            'bz2',
            'cairo',
            'calendar',
            'cmu_graphics',
            'code',
            'codecs',
            'codeop',
            'collections',
            'contextlib',
            'copy',
            'copyreg',
            'datetime',
            'decimal',
            'dis',
            'email',
            'encodings',
            'enum',
            'errno',
            'fcntl',
            'fnmatch',
            'functools',
            'genericpath',
            'hashlib',
            'heapq',
            'http',
            'importlib',
            'inspect',
            'io',
            'ipaddress',
            'itertools',
            'json',
            'keyword',
            'linecache',
            'locale',
            'lzma',
            'marshal',
            'math',
            'numbers',
            'opcode',
            'operator',
            'os',
            'platform',
            'posix',
            'posixpath',
            'pygame',
            'queue',
            'quopri',
            'random',
            're',
            'reprlib',
            'select',
            'selectors',
            'shutil',
            'signal',
            'site',
            'socket',
            'ssl',
            'stat',
            'string',
            'struct',
            'subprocess',
            'sys',
            'tempfile',
            'textwrap',
            'threading',
            'time',
            'token',
            'tokenize',
            'traceback',
            'types',
            'unicodedata',
            'urllib',
            'uuid',
            'warnings',
            'weakref',
            'zipimport',
            'zlib',
        ]
    }

    for directory in directories_to_look_for_shadowing:
        sibling_filenames = set(os.listdir(directory))
        overlap_filenames = sorted(list(sibling_filenames & disallowed_filenames))

        if overlap_filenames:
            raise Exception(f"""

        ******************************************************************************
        * The following files in {directory}
        * may prevent your program from running correctly. Please rename or remove
        * these files:
        *
        * {', '.join(overlap_filenames)}
        ******************************************************************************
        """)


check_for_shadowing()

from .libs import loader_util

loader_util.verify_support()

from cmu_graphics.cmu_graphics import (
    app,
    Arc,
    Circle,
    Group,
    Image,
    Label,
    Line,
    Oval,
    Polygon,
    Rect,
    RegularPolygon,
    Star,
    drawArc,
    drawCircle,
    drawImage,
    drawLabel,
    drawLine,
    drawOval,
    drawPolygon,
    drawRect,
    drawRegularPolygon,
    drawStar,
    ArcShape,
    CircleShape,
    ImageShape,
    LabelShape,
    LineShape,
    OvalShape,
    PolygonShape,
    RectShape,
    RegularPolygonShape,
    StarShape,
    Sound,
    gradient,
    rgb,
    almostEqual,
    rounded,
    round,
    dcos,
    dsin,
    onSteps,
    onKeyHolds,
    onKeyPresses,
    setLanguage,
    print,
    assertEqual,
    Robot,
    runApp,
    runAppWithScreens,
    setActiveScreen,
    getImageSize,
    pygameEvent,
    onStepEvent,
    onMainLoopEvent,
)

from cmu_graphics.utils import (
    angleTo,
    distance,
    fromPythonAngle,
    getPointInDir,
    makeList,
    pythonRound,
    toPythonAngle,
)

from random import (
    choice,
    random,
    randrange,
    seed,
)

from cmu_graphics.shape_logic import (
    TRANSLATED_GLOBALS,
    TRANSLATED_BOOLEANS,
    TRANSLATED_KEY_NAMES,
    accentCombinations,
    PILWrapper as CMUImage,
)

__all__ = TRANSLATED_GLOBALS['keys']
__all__.extend(
    [
        'setLanguage',
        'cmu_graphics',
        'runApp',
        'runAppWithScreens',
        'setActiveScreen',
        'getImageSize',
        'dcos',
        'dsin',
    ]
)
__all__.extend(
    [
        'drawArc',
        'ArcShape',
        'drawCircle',
        'CircleShape',
        'drawImage',
        'ImageShape',
        'drawLabel',
        'LabelShape',
        'drawLine',
        'LineShape',
        'drawOval',
        'OvalShape',
        'drawPolygon',
        'PolygonShape',
        'drawRect',
        'RectShape',
        'drawRegularPolygon',
        'RegularPolygonShape',
        'drawStar',
        'StarShape',
    ]
)

g = globals()
for language in TRANSLATED_GLOBALS:
    if language != 'keys':
        for en_name, trans_name in TRANSLATED_GLOBALS[language].items():
            if trans_name and trans_name != en_name and en_name in g:
                for accent_combination in accentCombinations(trans_name):
                    g[accent_combination] = g[en_name]
                    __all__.append(accent_combination)

for language in TRANSLATED_BOOLEANS:
    if language != 'keys':
        for en_name, trans_name in TRANSLATED_BOOLEANS[language].items():
            globals()[trans_name] = en_name == 'True'
            __all__.append(trans_name)

__all__ = list(set(__all__))



import decimal
import math
from . import shape_logic
from collections import defaultdict


def toDegrees(radians):
    return radians * 180 / math.pi


def toRadians(degrees):
    return degrees * math.pi / 180


def fromPythonAngle(radians):
    return (90 - toDegrees(radians)) % 360


def toPythonAngle(degrees):
    return (toRadians(90 - degrees)) % (2 * math.pi)


def intSin(degrees):
    if isinstance(degrees, float) and degrees.is_integer():
        degrees = int(degrees)
    if isinstance(degrees, int):
        degrees = degrees % 360
        if degrees == 0 or degrees == 180:
            return 0
        elif degrees == 90:
            return 1
        elif degrees == 270:
            return -1
    return math.sin(toRadians(degrees))


def intCos(degrees):
    if isinstance(degrees, float) and degrees.is_integer():
        degrees = int(degrees)
    if isinstance(degrees, int):
        degrees = degrees % 360
        if degrees == 90 or degrees == 270:
            return 0
        elif degrees == 0:
            return 1
        elif degrees == 180:
            return -1
    return math.cos(toRadians(degrees))


pythonRound = round


def round(*args):
    raise Exception(
        "Use our rounded(n) instead of Python 3's round(n)\n"
        "  Python 3's round(n) does not work as one might expect!\n"
        "  If you still want Python 3's round, use pythonRound"
    )


def rounded(d):
    sign = 1 if (d >= 0) else -1
    d = abs(d)
    n = int(d)
    if d - n >= 0.5:
        n += 1
    return sign * n


EPSILON = 10e-7


def almostEqual(x, y, epsilon=EPSILON):
    return abs(x - y) <= epsilon


def makeList(rows, cols, value=None):
    if rows < 0 or cols < 0:
        raise Exception('Both rows and cols must be >= 0')
    return [[value for _ in range(cols)] for _ in range(rows)]


def getPointInDir(x1, y1, degrees, d):
    A = toPythonAngle(degrees)
    return [x1 + d * math.cos(A), y1 - d * math.sin(A)]


def angleTo(x1, y1, x2, y2):
    dx = x2 - x1
    dy = y2 - y1
    return fromPythonAngle(math.atan2(-dy, dx))  # use -dy since up is down


def roundHalfUp(d):
    # Round to nearest with ties going away from zero.
    rounding = decimal.ROUND_HALF_UP
    return int(decimal.Decimal(d).to_integral_value(rounding=rounding))


def internalError(err):
    raise Exception('Internal Error: {err}'.format(err=err))


def polygonContainsPoint(pts, px, py):
    # based on: https://github.com/mathigon/fermat.js/blob/master/src/geometry.js
    n = len(pts)
    inside = False
    for i in range(n):
        q1 = pts[i]
        q2 = pts[(i + 1) % n]
        q1x = q1[0]
        q1y = q1[1]
        q2x = q2[0]
        q2y = q2[1]
        if distanceToLineSegment2(px, py, q1x, q1y, q2x, q2y) < 0.0002:
            return True
        x = (q1y > py) != (q2y > py)
        if q2y - q1y == 0:
            y = True
        else:
            y = px < (q2x - q1x) * (py - q1y) / (q2y - q1y) + q1x
        if x and y:
            inside = not inside
    return inside


def pointNearPolygonBorder(pts, x, y, d):
    # does not check if the polygon contains the point!
    d2 = d**2
    n = len(pts)
    for i in range(n):
        p1 = pts[i]
        p2 = pts[(i + 1) % n]
        x1 = p1[0]
        y1 = p1[1]
        x2 = p2[0]
        y2 = p2[1]
        if distanceToLineSegment2(x, y, x1, y1, x2, y2) <= d2:
            return True
    return False


def distance(x1, y1, x2, y2):
    return math.sqrt((x2 - x1) ** 2 + (y2 - y1) ** 2)


def distance2(x1, y1, x2, y2):
    return (x2 - x1) ** 2 + (y2 - y1) ** 2


def distanceToLineSegment2(x, y, x1, y1, x2, y2):
    # return square of distance from (x,y) to ((x1,y1),(x2,y2))
    # inspired by https://stackoverflow.com/questions/849211/shortest-distance-between-a-point-and-a-line-segment/24913403
    l2 = distance2(x1, y1, x2, y2)
    if l2 == 0:
        return distance(x, y, x1, y1)
    t = ((x - x1) * (x2 - x1) + (y - y1) * (y2 - y1)) / l2
    t = max(0, min(1, t))
    return distance2(x, y, x1 + t * (x2 - x1), y1 + t * (y2 - y1))


def edgesIntersect(edges1, edges2):
    ADD = True
    REMOVE = False

    x_to_events = defaultdict(list)
    for shape, edges in ((1, edges1), (2, edges2)):
        for edge in edges:
            # Assumes that x1 <= x2 for all edges
            (x1, _, x2, _) = edge
            x_to_events[x1].append((shape, ADD, edge))
            x_to_events[x2].append((shape, REMOVE, edge))

    active_edges1 = set()
    active_edges2 = set()

    # Here we're looping over a list of all line segments' start and ends points,
    # which is sorted by the points' x values.
    #
    # We're keeping track of which line segments are "active" as we go
    # (which line segments intersect with the line x=x for the current x value).
    #
    # When we encounter a new line segment, we see if it intersects with any
    # active line segments in the other shape.
    for _, events in sorted(x_to_events.items(), key=lambda item: item[0]):
        for shape, event_type, edge1 in events:
            my_active_edges, other_active_edges = (
                (active_edges1, active_edges2)
                if shape == 1
                else (active_edges2, active_edges1)
            )
            if event_type == ADD:
                for edge2 in other_active_edges:
                    if segmentsIntersect(*edge1, *edge2):
                        return True
                my_active_edges.add(edge1)
            else:
                my_active_edges.remove(edge1)

    return False


def segmentsIntersect(x1, y1, x2, y2, x3, y3, x4, y4):
    dxa = x2 - x1
    dya = y2 - y1
    dxb = x4 - x3
    dyb = y4 - y3
    s = (
        math.inf
        if (-dxb * dya + dxa * dyb) == 0
        else (-dya * (x1 - x3) + dxa * (y1 - y3)) / (-dxb * dya + dxa * dyb)
    )
    t = (
        math.inf
        if (-dxb * dya + dxa * dyb) == 0
        else (+dxb * (y1 - y3) - dyb * (x1 - x3)) / (-dxb * dya + dxa * dyb)
    )
    return s >= 0 and s <= 1 and t >= 0 and t <= 1


def isGroup(shape):
    if hasattr(shape, '_shape'):
        shape = shape._shape
    return isinstance(shape, shape_logic.Group)


def getChildShapes(shape):
    result = []
    if hasattr(shape, '_shape'):
        shape = shape._shape
    if isGroup(shape):
        for s in shape.children:
            result += getChildShapes(s)
    else:
        result = [shape]
    return result


def getPolygonArea(pts):
    A = 0
    for i in range(0, len(pts)):
        j = (i + 1) % len(pts)
        A += pts[i][0] * pts[j][1] - pts[j][0] * pts[i][1]
    return A / 2


def getPolygonCentroid(pts):
    A = getPolygonArea(pts)
    if A < 0.00001:
        # If the area of the polygon is small enough, average the points instead
        # of returning a value that is heavily influenced by floating point error
        sumX = 0
        sumY = 0
        for i in range(0, len(pts)):
            sumX += pts[i][0]
            sumY += pts[i][1]
        return [sumX / len(pts), sumY / len(pts)]
    cx, cy = 0, 0
    for i in range(0, len(pts)):
        j = (i + 1) % len(pts)
        term = pts[i][0] * pts[j][1] - pts[j][0] * pts[i][1]
        cx += (pts[i][0] + pts[j][0]) * term
        cy += (pts[i][1] + pts[j][1]) * term
    return [cx / (6 * A), cy / (6 * A)]


def rotatePoint(pt, degrees, cx, cy):
    [x, y] = pt
    cos = intCos(degrees)
    sin = intSin(degrees)
    return [
        cx + ((x - cx) * cos - (y - cy) * sin),
        cy + ((x - cx) * sin + (y - cy) * cos),
    ]


def rotatePoints(pts, degrees, cx, cy):
    return list(map(lambda pt: rotatePoint(pt, degrees, cx, cy), pts))


def getBoxDims(pts):
    if len(pts) == 0:
        internalError('getBoxDims: empty points list')

    xlo = xhi = pts[0][0]
    ylo = yhi = pts[0][1]
    for pt in pts:
        [x, y] = pt
        if x < xlo:
            xlo = x
        elif x > xhi:
            xhi = x
        if y < ylo:
            ylo = y
        elif y > yhi:
            yhi = y
    return {'left': xlo, 'top': ylo, 'width': xhi - xlo, 'height': yhi - ylo}


def flatten(a):
    out = []
    for elem in a:
        if isinstance(elem, list):
            out.extend(flatten(elem))
        else:
            out.append(elem)
    return out


def truncateIntegerFloats(n):
    if isinstance(n, float) and n.is_integer():
        return int(n)
    return n


def utilsRounded(n, precision=0):
    if isinstance(n, list) or isinstance(n, tuple):
        return list(map(lambda v: utilsRounded(v, precision), n))
    elif not (isinstance(n, int) or isinstance(n, float)):
        return n
    elif n < 0:
        return -utilsRounded(-n, precision)
    return truncateIntegerFloats(roundHalfUp(n * 10**precision) / 10**precision)


def tupleString(a):
    return '({s})'.format(s=', '.join(map(str, a)))


def roundedTupleString(a, precision=0):
    return tupleString(utilsRounded(a, precision))


def getArcPoints(
    cx, cy, width, height, startAngle=None, sweepAngle=None, sizeForN=None, isMvc=False
):
    # get points that approximate an oval
    a, b = width / 2, height / 2
    pts = []
    if sizeForN is None:
        sizeForN = (a + b) / 2
    if startAngle is None:
        startAngle = 0
    if sweepAngle is None:
        sweepAngle = 360
    else:
        pts.append([cx, cy])
    n = rounded(6 + 18 * sizeForN / 50)
    n = math.ceil(n / 4) * 4
    denominator = n if sweepAngle == 360 else n - 1
    startAngle = toRadians(startAngle) if isMvc else toRadians(90 - startAngle)
    sweepAngle = toRadians(sweepAngle)
    multiplyFactor = 1 if isMvc else -1
    for i in range(n):
        theta = startAngle + (multiplyFactor * (sweepAngle * i / denominator))
        x = cx + a * math.cos(theta)
        y = cy - b * math.sin(theta)
        pts.append([x, y])
    return pts


def isNumber(value):
    return isinstance(value, int) or isinstance(value, float)


def round6(value):
    return pythonRound((value + 0.00000001) * 1000000) / 1000000


def round2(value):
    return pythonRound((value + 0.001) * 100) / 100


def makePolygonPath(pts, ctx):
    ctx.new_path()
    if pts is None or len(pts) == 0:
        return
    lastPt = pts[-1]
    ctx.move_to(lastPt[0], lastPt[1])
    for pt in pts:
        ctx.line_to(pt[0], pt[1])


def getLinePoints(x1, y1, x2, y2, lineWidth):
    # 0. get angle of rotation clockwise past horizontal
    cx = (x1 + x2) / 2
    cy = (y1 + y2) / 2
    a = angleTo(x1, y1, x2, y2) - 90  # -90 so we're off the horizontal

    # 1. unrotate to horizontal, with p3 n the left, p4 on the right
    pts = [[x1, y1], [x2, y2]]
    pts = rotatePoints(pts, -a, cx, cy)
    x3 = pts[0][0]
    y3 = pts[0][1]
    x4 = pts[1][0]
    y4 = pts[1][1]  # y3 should be isClose to y4

    # 2. define bounding box around the unrotated-to-horizontal line
    s = lineWidth / 2
    pts = [[x3, y3 - s], [x4, y3 - s], [x4, y4 + s], [x3, y4 + s]]
    return rotatePoints(pts, a, cx, cy)


def getRegularPolygonPoints(cx, cy, r, points, rotateAngle):
    pts = [[cx, cy - r]]
    dtheta = 360 / points
    for i in range(1, points):
        [x, y] = getPointInDir(cx, cy, i * dtheta, r)
        pts.append([x, y])
    if rotateAngle:
        pts = rotatePoints(pts, rotateAngle, cx, cy)
    return pts


def getDefaultRoundness(points):
    return 38.196601125 if points < 6 else 57.735026919


def getStarPoints(cx, cy, r, points, roundness, rotateAngle):
    if roundness is None:
        roundness = getDefaultRoundness(points)
    if roundness < 5:
        roundness = 5
    innerR = r * roundness / 100
    pts = [[cx, cy - r]]
    dtheta = 360 / points
    for i in range(points):
        if i > 0:
            [x, y] = getPointInDir(cx, cy, i * dtheta, r)
            pts.append([x, y])
        [x, y] = getPointInDir(cx, cy, i * dtheta + dtheta / 2, innerR)
        pts.append([x, y])
    if rotateAngle:
        pts = rotatePoints(pts, rotateAngle, cx, cy)
    return pts


def convertLabelValue(value):
    if callable(value):
        return '<function>'
    return str(value)


def min_or_inf(L):
    if len(L) == 0:
        return math.inf
    return min(L)



import sys
from cmu_graphics import *

# The base URL for the API we are using: metaweather
API_BASE = 'https://www.metaweather.com'

def celciusToFarenheit(degrees):
    return ((degrees / 5) * 9) + 32

# Given a name of a city, this function returns information about the weather
# in that city, and information about the city itself
def getAPIData(city):
    # First query the API to get information about the city given its name
    cityData = requests.get(
        API_BASE + '/api/location/search/?query=' + city
    ).json()[0]

    # Extract the Where On Earth Identifier from the city information
    cityWoeid = cityData['woeid']

    # Then query the API using the woeid to get its weather
    weatherData = requests.get(
        API_BASE + '/api/location/' + str(cityWoeid) + '/'
    ).json()['consolidated_weather'][0]

    return weatherData, cityData

# Given a city, queries the metaweather API to find its weather information,
# then draws that weather to the canvas
def drawWeather(city):
    app.group.clear()

    weatherData, cityData = getAPIData(city)

    # Draws the title "Weather in City: Rainy"
    Label('Weather in ', 200, 30, size=40)
    Label(cityData['title'], 200, 75, size=40)
    Label(weatherData['weather_state_name'], 200, 130, size=30)

    # Given the weather state (sunny, cloudy, windy, etc)
    # Get an image from metawetaher representing that weather state
    weather_state = weatherData['weather_state_abbr']
    imageURL = API_BASE + '/static/img/weather/png/'+ weather_state + '.png'
    Image(imageURL, 200, 230, align='center', width=150, height=150)

    # Draws the temperature in Farenheit and the wind direction
    # The temperature returned from metaweather is in Celcius, so we have
    # to convert it first
    farenheitTemp = celciusToFarenheit(weatherData['the_temp'])
    Label(str(pythonRound(farenheitTemp, 2)) + '°F',
          25, 350, align='left', size=30)
    Label('Wind: ' + weatherData['wind_direction_compass'],
          375, 350, align='right', size=30)

    # Make the back button visible so we can return to the search page
    app.backButton.visible = True

def onMousePress(mouseX, mouseY):
    # If the back button is visible (which means we're displaying the weather
    # for a city right now), and we click it, go back to the search page
    if app.backButton.visible and app.backButton.hits(mouseX, mouseY):
        app.backButton.visible = False
        app.group.clear()
        drawSearchPage()

def drawSearchPage():
    Label('Press S to search for a city.', 200, 200, size=25)

# Search for a city when we press the s key, and try to draw its weather
# If we can't find the city, print an error message
def onKeyPress(key):
    if key.lower() == 's':
        city = app.getTextInput('City Name:')
        try:
            drawWeather(city)
        except:
            Label('City not found. Try another!', 200, 250, size=20)

# A helper function for drawing text centered on the screen
def drawText(linesList):
    lineHeight = 30
    # lineY starts out at the center (200) minus half the height of our
    # whole block of text
    lineY = 200 - ((len(linesList) * lineHeight) // 2)
    for line in linesList:
        # Create a label for each line, and move lineY down so the next
        # line is drawn lower
        Label(line, 200, lineY, align='center', size=20)
        lineY += lineHeight

# This function loads the third party library requests, and adds
# it to the global namespace so we can use it in other functions.
def loadModules():
    global requests
    import requests

# Try to load the necessary third party libraries, create our back button, and
# start the application. If loading the libraries fails, print a message about
# installation
def start():
    try:
        loadModules()
        # The visibility of the back button changes when we draw the weather,
        # and it is accessed in the mouse press handler, so it needs to be
        # part of the app
        app.backButton = Group(
            Line(30, 30, 80, 30, lineWidth=10),
            RegularPolygon(30, 30, 15, 3, rotateAngle=-90),
            visible=False
        )
        drawSearchPage()
    except:
        drawText([
            'This program requires "requests".',
            'This is third party library which can',
            'be installed by running the command',
            'printed to the console in the application',
            '"Command Prompt".' if sys.platform == 'win32' else '"Terminal".',
        ])
        if sys.platform == 'win32':
            print('"%s" -m pip install requests' % sys.executable)
        else:
            print('sudo "%s" -m pip install requests' % sys.executable)
        app.mode = 'helpText'

start()

cmu_graphics.run()



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open("cmu_graphics/meta/version.txt", "r", encoding="utf-8") as f:
    version = f.read().strip()

setuptools.setup(
    name="cmu-graphics",
    version=version,
    author="Austin Schick",
    author_email="aschick@andrew.cmu.edu",
    description="The graphics framework used by CMU CS Academy, geared toward beginner CS students.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://academy.cs.cmu.edu/",
    project_urls={
        "Documentation": "https://academy.cs.cmu.edu/docs",
    },
    license="BSD 3-Clause License",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "pygame>=2",
        "pycairo>=1.20",
    ],
    packages=["cmu_graphics", "cmu_graphics.samples", "cmu_graphics.libs"],
    package_data={"cmu_graphics": ["meta/version.txt", "libs/cacert.pem"]},
    python_requires=">=3.6, <4",
)
